#include <direct.h>
#include <string>
#include "common/IFileStream.h"

#include "Options.h"
#include "Inject.h"

IDebugLog	gLog("nvse_loader.log");

enum
{
	kProcType_Steam,
	kProcType_Normal,

	kProcType_Unknown
};

static bool IdentifyEXE(const char * procName, bool isEditor, std::string * dllSuffix, ProcHookInfo * hookInfo);
static bool InjectDLL(PROCESS_INFORMATION * info, const char * dllPath, ProcHookInfo * hookInfo);
static void PrintError(const char * fmt, ...);

int main(int argc, char ** argv)
{
	gLog.SetPrintLevel(IDebugLog::kLevel_FatalError);
	gLog.SetLogLevel(IDebugLog::kLevel_DebugMessage);

	if(!g_options.Read(argc, argv))
	{
		PrintError("Couldn't read arguments.");
		g_options.PrintUsage();

		return -1;
	}

	if(g_options.m_optionsOnly)
	{
		g_options.PrintUsage();
		return 0;
	}

	if(g_options.m_verbose)
		gLog.SetPrintLevel(IDebugLog::kLevel_VerboseMessage);

	if(g_options.m_launchCS)
		_MESSAGE("launching editor");

	// create the process
	bool	dllHasFullPath = false;

	const char	* procName = g_options.m_launchCS ? "GECK.exe" : "FalloutNV.exe";
	const char	* baseDllName = g_options.m_launchCS ? "nvse_editor" : "nvse";

	char	currentWorkingDirectory[4096];
	ASSERT(_getcwd(currentWorkingDirectory, sizeof(currentWorkingDirectory)));

	std::string procPath = std::string(currentWorkingDirectory) + "\\" + std::string(procName);

	if(g_options.m_altEXE.size())
	{
		procPath = g_options.m_altEXE;
		_MESSAGE("launching alternate exe (%s)", procPath.c_str());
	}
	else if(!g_options.m_launchCS)
	{
		IFileStream	fileCheck;
		if(!fileCheck.Open(procPath.c_str()))
		{
#if 0
			// ### not sure what the ng version is called
			const char * kNoGoreEXEName = "falloutnvng.exe";

			_MESSAGE("couldn't find %s, trying %s", procName, kNoGoreEXEName);

			procName = kNoGoreEXEName;
			procPath = std::string(currentWorkingDirectory) + "\\" + std::string(procName);

#else
			PrintError("Couldn't find %s.", procName);
			return -1;
#endif
		}
	}

	_MESSAGE("launching: %s (%s)", procName, procPath.c_str());

	if(g_options.m_altDLL.size())
	{
		baseDllName = g_options.m_altDLL.c_str();
		_MESSAGE("launching alternate dll (%s)", baseDllName);

		dllHasFullPath = true;
	}

	std::string		dllSuffix;
	ProcHookInfo	procHookInfo;

	if(!IdentifyEXE(procPath.c_str(), g_options.m_launchCS, &dllSuffix, &procHookInfo))
	{
		_ERROR("unknown exe");
		return -1;
	}

	_MESSAGE("hook call addr = %08X", procHookInfo.hookCallAddr);
	_MESSAGE("load lib addr = %08X", procHookInfo.loadLibAddr);

	if(g_options.m_crcOnly)
		return 0;

	// build dll path
	std::string	dllPath;
	if(dllHasFullPath)
	{
		dllPath = baseDllName;
	}
	else
	{
		dllPath = std::string(currentWorkingDirectory) + "\\" + baseDllName + "_" + dllSuffix + ".dll";
	}

	_MESSAGE("dll = %s", dllPath.c_str());

	// check to make sure the dll exists
	{
		IFileStream	tempFile;

		if(!tempFile.Open(dllPath.c_str()))
		{
			PrintError("Couldn't find NVSE DLL (%s). Please make sure you have installed NVSE correctly and are running it from your Fallout folder.", dllPath.c_str());
			return -1;
		}
	}

	STARTUPINFO			startupInfo = { 0 };
	PROCESS_INFORMATION	procInfo = { 0 };

	startupInfo.cb = sizeof(startupInfo);

	if(!CreateProcess(
		procPath.c_str(),
		NULL,	// no args
		NULL,	// default process security
		NULL,	// default thread security
		TRUE,	// don't inherit handles
		CREATE_SUSPENDED,
		NULL,	// no new environment
		NULL,	// no new cwd
		&startupInfo, &procInfo))
	{
		PrintError("Launching %s failed (%d).", procPath.c_str(), GetLastError());
		return -1;
	}

	bool	injectionSucceeded = false;

	switch(procHookInfo.procType)
	{
		case kProcType_Steam:
			injectionSucceeded = true;
			break;

		case kProcType_Normal:
			if(InjectDLL(&procInfo, dllPath.c_str(), &procHookInfo))
			{
				injectionSucceeded = true;

				if(g_options.m_launchCS && g_options.m_loadScriptErrorHook)
				{
					_MESSAGE("loading script error hook");

					//	this hooks 005C51ED with
					//	
					//	int PrintScriptErr(int * outErr, const char * fmt, va_list args)
					//	{
					//		char	buf[0x1300];
					//		
					//		if(fmt)
					//		{
					//			_vsprintf(buf, fmt, args);
					//			if(strstr(buf, "SCRIPTS: Script \'DialogueCompileAndRun\'"))
					//			{
					//				MessageBox(NULL, buf, NULL, MB_OK);
					//			}
					//		}
					//		
					//		return 0;	// should probably be the result of _vsprintf
					//	}
					//	
					//	need to contact the person who originally wrote this to see if we can integrate it

					if(!InjectDLL(&procInfo, "NVGeckExt.dll", &procHookInfo))
					{
						_ERROR("couldn't inject NVGeckExt.dll");
					}
				}
			}
			break;

		default:
			HALT("impossible");
	}

	if(!injectionSucceeded)
	{
		PrintError("Couldn't inject DLL.");

		_ERROR("terminating process");

		TerminateProcess(procInfo.hProcess, 0);
	}
	else
	{
		_MESSAGE("launching");

		ResumeThread(procInfo.hThread);

		if(g_options.m_waitForClose)
			WaitForSingleObject(procInfo.hProcess, INFINITE);
	}

	CloseHandle(procInfo.hProcess);
	CloseHandle(procInfo.hThread);

	return 0;
}

bool GetFileVersion(const char * path, VS_FIXEDFILEINFO * info)
{
	bool result = false;

	UInt32	versionSize = GetFileVersionInfoSize(path, NULL);
	if(!versionSize)
	{
		_ERROR("GetFileVersionInfoSize failed (%08X)", GetLastError());
		return false;
	}

	UInt8	* versionBuf = new UInt8[versionSize];
	if(versionBuf)
	{
		if(GetFileVersionInfo(path, NULL, versionSize, versionBuf))
		{
			VS_FIXEDFILEINFO	* retrievedInfo = NULL;
			UInt32				realVersionSize = sizeof(VS_FIXEDFILEINFO);

			if(VerQueryValue(versionBuf, "\\", (void **)&retrievedInfo, (PUINT)&realVersionSize) && retrievedInfo)
			{
				*info = *retrievedInfo;
				result = true;
			}
			else
			{
				_ERROR("VerQueryValue failed (%08X)", GetLastError());
			}
		}
		else
		{
			_ERROR("GetFileVersionInfo failed (%08X)", GetLastError());
		}

		delete [] versionBuf;
	}

	return result;
}

bool GetFileVersionNumber(const char * path, UInt64 * out)
{
	VS_FIXEDFILEINFO	versionInfo;
	if(!GetFileVersion(path, &versionInfo))
	{
		return false;
	}

	_MESSAGE("dwSignature = %08X", versionInfo.dwSignature);
	_MESSAGE("dwStrucVersion = %08X", versionInfo.dwStrucVersion);
	_MESSAGE("dwFileVersionMS = %08X", versionInfo.dwFileVersionMS);
	_MESSAGE("dwFileVersionLS = %08X", versionInfo.dwFileVersionLS);
	_MESSAGE("dwProductVersionMS = %08X", versionInfo.dwProductVersionMS);
	_MESSAGE("dwProductVersionLS = %08X", versionInfo.dwProductVersionLS);
	_MESSAGE("dwFileFlagsMask = %08X", versionInfo.dwFileFlagsMask);
	_MESSAGE("dwFileFlags = %08X", versionInfo.dwFileFlags);
	_MESSAGE("dwFileOS = %08X", versionInfo.dwFileOS);
	_MESSAGE("dwFileType = %08X", versionInfo.dwFileType);
	_MESSAGE("dwFileSubtype = %08X", versionInfo.dwFileSubtype);
	_MESSAGE("dwFileDateMS = %08X", versionInfo.dwFileDateMS);
	_MESSAGE("dwFileDateLS = %08X", versionInfo.dwFileDateLS);

	UInt64 version = (((UInt64)versionInfo.dwFileVersionMS) << 32) | versionInfo.dwFileVersionLS;

	*out = version;

	return true;
}

bool IsSteamImage(UInt8 * base)
{
	// ### yes this can overrun if the file is corrupt
	// ### if the file is corrupt then GetFileVersion probably died earlier

	IMAGE_DOS_HEADER		* dosHeader = (IMAGE_DOS_HEADER *)base;
	IMAGE_NT_HEADERS		* ntHeader = (IMAGE_NT_HEADERS *)(base + dosHeader->e_lfanew);
	IMAGE_SECTION_HEADER	* sectionHeader = IMAGE_FIRST_SECTION(ntHeader);

	for(UInt32 i = 0; i < ntHeader->FileHeader.NumberOfSections; i++)
	{
		IMAGE_SECTION_HEADER	* section = &sectionHeader[i];

		if(!strcmp((const char *)section->Name, ".bind"))
		{
			return true;
		}
	}

	return false;
}

// steam EXE will have the .bind section
bool IsSteamEXE(const char * path, bool * out)
{
	// open and map the file in to memory
	HANDLE	file = CreateFile(path, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if(file == INVALID_HANDLE_VALUE)
	{
		_ERROR("IsSteamEXE: couldn't open file (%d)", GetLastError());
		return false;
	}

	bool result = false;

	HANDLE	mapping = CreateFileMapping(file, NULL, PAGE_READONLY, 0, 0, NULL);
	if(mapping)
	{
		UInt8	* fileBase = (UInt8 *)MapViewOfFile(mapping, FILE_MAP_READ, 0, 0, 0);
		if(fileBase)
		{
			*out = IsSteamImage(fileBase);
			result = true;

			UnmapViewOfFile(fileBase);
		}
		else
		{
			_ERROR("IsSteamEXE: couldn't map file (%d)", GetLastError());
		}

		CloseHandle(mapping);
	}
	else
	{
		_ERROR("IsSteamEXE: couldn't create file mapping (%d)", GetLastError());
	}

	CloseHandle(file);

	return result;
}

bool IdentifyEXE(const char * procName, bool isEditor, std::string * dllSuffix, ProcHookInfo * hookInfo)
{
	UInt64	version;
	bool	isSteam;

	// check file version
	if(!GetFileVersionNumber(procName, &version))
	{
		PrintError("Couldn't retrieve EXE version information.");
		return false;
	}

	_MESSAGE("version = %016I64X", version);

	// check for Steam EXE
	if(!IsSteamEXE(procName, &isSteam))
	{
		PrintError("Failed to identify Steam EXE.");
		return false;
	}

	// ### how to tell the difference between nogore versions and standard without a global checksum?
	// ### since we're mapping the exe to check for steam anyway, checksum the .text segment maybe?

	_MESSAGE(isSteam ? "steam exe" : "normal exe");

	hookInfo->version = version;
	hookInfo->procType = isSteam ? kProcType_Steam : kProcType_Normal;

	bool result = true;

	if(isEditor)
	{
		switch(version)
		{
			case 0x0001000100000106:	// 1.1.0.262 (original release)
				hookInfo->hookCallAddr = 0x00C61BA1;
				hookInfo->loadLibAddr = 0x00D2218C;
				*dllSuffix = "1_1";
				break;

			default:
				PrintError("You have an unknown version of the CS. Please check http://nvse.silverlock.org to make sure you're using the latest version of NVSE. (version = %016I64X)", version);
				result = false;
				break;
		}
	}
	else
	{
		if(version < 0x000100010001010F)
		{
			PrintError("Please update to the latest version of Fallout.");
		}
		else if(version == 0x000100010001010F)
		{
			// 1.1.1.271

			hookInfo->hookCallAddr = 0x00EC356B;
			hookInfo->loadLibAddr = 0x00FD60B0;
			*dllSuffix = "1_0";
		}
		else
		{
			PrintError("You have an unknown version of Fallout. Please check http://nvse.silverlock.org to make sure you're using the latest version of NVSE. (version = %016I64X)", version);
			result = false;
		}
	}

	return result;
}

static bool InjectDLL(PROCESS_INFORMATION * info, const char * dllPath, ProcHookInfo * hookInfo)
{
	bool	result = false;

	// wrap DLL injection in SEH, if it crashes print a message
	__try {
		result = DoInjectDLL(info, dllPath, hookInfo);
	}
	__except(EXCEPTION_EXECUTE_HANDLER)
	{
		PrintError("DLL injection failed. In most cases, this is caused by an overly paranoid software firewall or antivirus package. Disabling either of these may solve the problem.");
		result = false;
	}

	return result;
}

static void PrintError(const char * fmt, ...)
{
	va_list	args;

	va_start(args, fmt);

	gLog.Log(IDebugLog::kLevel_FatalError, fmt, args);

	char	buf[4096];
	vsprintf_s(buf, sizeof(buf), fmt, args);

	MessageBox(NULL, buf, "NVSE Loader", MB_OK | MB_ICONEXCLAMATION);

	va_end(args);
}
